(* Utility Declarations *)
(* Copyright (c) 2005, Rodrigo Ferreira *)

Require Import Classical.

Set Implicit Arguments.

(* Type Option *)
Inductive optionT (A : Type) : Type :=
  | SomeT (a : A)
  | NoneT.
Implicit Arguments NoneT [A].

Reserved Notation "? A" (at level 70, right associativity).
Notation "? A" := (optionT A) : type_scope.

(* Type Product *)
Notation "x * y" := (prodT x y) : type_scope.
Notation "( x , y , .. , z )" := (pairT .. (pairT x y) .. z) : core_scope.

Lemma trv1 : forall a, a <> true -> a = false.
Proof.
intros a Ht.
induction a.
elim Ht.
trivial.
trivial.
Qed.

(* Type Function *)
Section TypeFun.

Variables A B : Type.
Variable beq_A : A -> A -> bool.

Definition def (b : B) : A -> B :=
  fun a' => b.

Definition ext (f : A -> B) (a : A) (b : B) :=
  fun a' => if beq_A a' a then b else f a'.

End TypeFun.

(* Type Map *)
Section TypeMap.

Variables A B : Type.
Variable beq_A : A -> A -> bool.

Definition emp : A ->? B :=
  fun a' => NoneT.

Definition has (m : A ->? B) (a : A) :=
  exists b, m a = SomeT b.

Definition get (m : A ->? B) (a : A) (b : B) :=
  m a = SomeT b.

Definition ins (m : A ->? B) (a : A) (b : B) :=
  fun a' => if beq_A a' a then SomeT b else m a'.

Definition del (m : A ->? B) (a : A) :=
  fun a' => if beq_A a' a then NoneT else m a'.

Lemma has_ins_intro : forall (m : A ->? B) (a a' : A) (b : B), has (ins m a b) a' -> beq_A a' a = true \/ (beq_A a' a = false /\ has m a').
Proof.
intros m a a' b Hhas.
cut (beq_A a' a = true \/ beq_A a' a <> true).
intros Ht.
elim Ht.
clear Ht.
intro Heq.
left.
exact Heq.
clear Ht.
intro Hne.
right.
split.
apply trv1.
exact Hne.
unfold has in Hhas.
unfold has.
elim Hhas.
intros b' Hins.
exists b'.
rewrite <- Hins.
unfold ins.
rewrite -> (trv1 Hne).
trivial.
apply classic.
Qed.

Lemma has_ins_elimi : forall (m : A ->? B) (a a' : A) (b : B), beq_A a' a = true \/ (beq_A a' a = false /\ has m a') -> has (ins m a b) a'.
Proof.
intros m a a' b Ht.
elim Ht.
clear Ht.
intro Heq.
unfold ins.
unfold has.
exists b.
rewrite -> Heq.
trivial.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hhas.
unfold ins.
unfold has.
unfold has in Hhas.
elim Hhas.
intros b' Ht.
exists b'.
rewrite -> Hne.
exact Ht.
Qed.

Lemma get_ins_intro : forall (m : A ->? B) (a a' : A) (b b' : B), get (ins m a b) a' b' -> (beq_A a' a = true /\ b = b') \/ (beq_A a' a = false /\ get m a' b').
Proof.
intros m a a' b b' Hget.
cut (beq_A a' a = true \/ beq_A a' a <> true).
intros Ht.
elim Ht.
clear Ht.
intro Heq.
left.
split.
exact Heq.
unfold ins in Hget.
unfold get in Hget.
rewrite -> Heq in Hget.
inversion Hget.
rename H0 into Ht.
trivial.
clear Ht.
intro Hne.
right.
split.
apply trv1.
exact Hne.
unfold get in Hget.
unfold get.
rewrite <- Hget.
unfold ins.
rewrite -> (trv1 Hne).
trivial.
apply classic.
Qed.

Lemma get_ins_elimi : forall (m : A ->? B) (a a' : A) (b b' : B), (beq_A a' a = true /\ b = b') \/ (beq_A a' a = false /\ get m a' b') -> get (ins m a b) a' b'.
Proof.
intros m a a' b b' Ht.
elim Ht.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Heq Heq'.
unfold ins.
unfold get.
rewrite -> Heq.
rewrite <- Heq'.
trivial.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hget.
unfold ins.
unfold get.
unfold get in Hget.
rewrite -> Hne.
exact Hget.
Qed.

Lemma has_del_intro : forall (m : A ->? B) (a a' : A), has (del m a) a' -> beq_A a' a = false /\ has m a'.
Proof.
intros m a a' Hhas.
unfold has in Hhas.
elim Hhas.
clear Hhas.
intros b Hdel.
unfold del in Hdel.
cut (beq_A a' a = true \/ beq_A a' a <> true).
intro Ht.
elim Ht.
clear Ht.
intro Heq.
rewrite -> Heq in Hdel.
discriminate.
clear Ht.
intro Hne.
split.
apply trv1.
exact Hne.
rewrite -> (trv1 Hne) in Hdel.
unfold has.
exists b.
exact Hdel.
apply classic.
Qed.

Lemma has_del_elimi : forall (m : A ->? B) (a a' : A), beq_A a' a = false /\ has m a' -> has (del m a) a'.
Proof.
intros m a a' Ht.
elim Ht.
clear Ht.
intros Hne Hhas.
unfold del.
unfold has.
unfold has in Hhas.
elim Hhas.
intros b Ht.
exists b.
rewrite -> Hne.
exact Ht.
Qed.

Lemma get_del_intro : forall (m : A ->? B) (a a' : A) (b' : B), get (del m a) a' b' -> beq_A a' a = false /\ get m a' b'.
Proof.
intros m a a' b' Hget.
unfold get in Hget.
unfold del in Hget.
cut (beq_A a' a = true \/ beq_A a' a <> true).
intro Ht.
elim Ht.
clear Ht.
intro Heq.
rewrite -> Heq in Hget.
discriminate.
clear Ht.
intro Hne.
split.
apply trv1.
exact Hne.
rewrite -> (trv1 Hne) in Hget.
unfold get.
exact Hget.
apply classic.
Qed.

Lemma get_del_elimi : forall (m : A ->? B) (a a' : A) (b' : B), beq_A a' a = false /\ get m a' b' -> get (del m a) a' b'.
Proof.
intros m a a' b' Ht.
elim Ht.
clear Ht.
intros Hne Hget.
unfold del.
unfold get.
unfold get in Hget.
rewrite -> Hne.
exact Hget.
Qed.

End TypeMap.

