(* The LOOP program *)
(* Copyright (c) 2005, FLINT Group, Yale University *)

Require Import Classical.

Require Import Util.
(* Require Import HiMach. *)
Require Import SoundSafe.
Require Import Omega.
Require Import EqNat.
Require Import ThMach.
Require Import natutil.

Set Implicit Arguments.

(* code heap labels *)
Notation L_main := 1 (only parsing).
Notation L_loop := 2 (only parsing).
Notation L_cont := 3 (only parsing).
Notation L_chld := 4 (only parsing).

(* variables *)
Notation V_data := 0 (only parsing).
Notation V_length := 100 (only parsing).

(* definition of code heap *)
Definition C_main : block :=
  seq (movi r0 0)
 (seq (movi r1 1)
 (seq (movi r2 V_length)
      (jd L_loop)
 )).

Definition C_loop : block :=
  seq (beq r0 r2 L_cont)
 (seq (yield)
 (seq (st r0 V_data r0)
 (seq (yield)
 (seq (fork L_chld r0)
 (seq (yield)
 (seq (add r0 r0 r1)
 (seq (yield)
      (jd L_loop)
 ))))))).

Definition C_cont : block :=
  seq (yield) (jd L_cont).

Definition C_chld : block :=
  seq (movi r0 V_data)
 (seq (yield)
 (seq (add r0 r0 ta)
 (seq (yield)
 (seq (ld r1 r0 0)
 (seq (yield)
 (seq (movi r2 1)
 (seq (yield)
 (seq (add r1 r1 r2)
 (seq (yield)
 (seq (st r0 0 r1)
      (exit)
 )))))))))).

Definition CdHeap : code :=
  fun (f : label) =>
    if (beq_w f L_main)
    then SomeT C_main
    else if (beq_w f L_loop)
	 then SomeT C_loop
	 else if (beq_w f L_cont)
         then SomeT C_cont
         else if (beq_w f L_chld)
              then SomeT C_chld
              else NoneT.


(* code specifications *)

Definition Inv : INV := fun M : heap => True.

Definition T_main_p : precond := fun s : state => True.

Definition T_main_A : guarantee := 
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> M (V_data + i) = M' (V_data + i).

Definition T_main_G : guarantee :=
  fun (s : state) (M' : heap) => True.

Definition T_loop_p : precond :=
  fun s : state =>
    let (R,M) := s
    in (R r0 >= 0) /\ (R r0 <= V_length) /\ R r1 = 1 /\ R r2 = V_length.

Definition T_loop_A : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> (i >= R r0) -> M (V_data + i) = M' (V_data + i).

Definition T_loop_G : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> (i < R r0) -> M (V_data + i) = M' (V_data + i).

Definition T_fork_A : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> (i > R r0) -> M (V_data + i) = M' (V_data + i).

Definition T_fork_G : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> (i <= R r0) -> M (V_data + i) = M' (V_data + i).

Definition T_cont_p : precond := 
  fun s : state => 
    let (R,M) := s
    in  R r0 = V_length.

Definition T_cont_A : guarantee :=
  fun (s : state) (M' : heap) => True.

Definition T_cont_G : guarantee := 
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> M (V_data + i) = M' (V_data + i).

Definition T_chld_p : precond := 
  fun s : state =>
    let (R,M) := s
    in (R ta >= 0) /\ (R ta < V_length).

Definition T_chld_A : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  M (R ta) = M' (R ta).

Definition T_chld_G : guarantee := 
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  forall i : nat, (i >= 0) -> (i < V_length) -> (i <> R ta) -> M (V_data + i) = M' (V_data + i).

Definition T_G_id : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in M = M'.

Definition Psi : cspec :=
  fun f : label => 
    if (beq_w f L_main)
    then SomeT (T_main_p, T_main_G, T_main_A, T_main_G)
    else if (beq_w f L_loop)
         then SomeT (T_loop_p, T_loop_G, T_loop_A, T_loop_G)
	 else if (beq_w f L_cont)
	      then SomeT (T_cont_p, T_cont_G, T_cont_A, T_cont_G)
	      else if (beq_w f L_chld)
		   then SomeT (T_chld_p, T_chld_G, T_chld_A, T_chld_G)
		   else NoneT.


(* code verification *)
Lemma wf_main : I_infer Inv Psi (T_main_p, T_main_G, T_main_A, T_main_G) C_main.
Proof.
unfold C_main.

(* movi r0 0 *)
apply I_infer_movi with (fun s:state => let (R,M) := s in R r0 = 0) T_main_G T_main_A T_main_G.
discriminate.
discriminate.
intros.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
auto.
auto.
auto.

(* movi r1 1 *)
apply I_infer_movi with (fun s:state => let (R,M) := s in R r0 = 0 /\ R r1 = 1) T_main_G T_main_A T_main_G.
discriminate.
discriminate.
intros.
split.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
auto.
auto.
auto.

(* movi r2 100 *)
apply I_infer_movi with (fun s:state => let (R,M) := s in R r0 = 0 /\ R r1 = 1 /\ R r2 = V_length) T_main_G T_main_A T_main_G.
discriminate.
discriminate.
intros.
split.
unfold ext_r.
unfold ext.
unfold beq_r.
inversion_clear H0.
auto.
split.
unfold ext_r.
unfold ext.
unfold beq_r.
inversion_clear H0.
auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
auto.
auto.
auto.

(* jd L_loop *)
apply I_infer_jd with T_loop_p T_loop_G T_loop_A T_loop_G.
unfold T_loop_p.
intros.
omega.
intros.
unfold T_main_G.
auto.
unfold T_main_A.
unfold T_loop_A.
intros.
apply H1; auto.
unfold T_main_G.
auto.
unfold Psi.
unfold get.
auto.
Qed.

Lemma wf_loop : I_infer Inv Psi (T_loop_p, T_loop_G, T_loop_A, T_loop_G) C_loop.
Proof.
unfold C_loop.
(* beq r0 r2 L_cont *)
apply I_infer_beq with 
  (fun s : state => let (R,M) := s in (R r0 >= 0) /\ (R r0 < V_length) /\ R r1 = 1 /\ R r2 = V_length) T_cont_p 
  T_G_id T_cont_G T_loop_A T_cont_A T_loop_G T_cont_G.
intros.
(* inversion_clear H1.*)
unfold T_loop_p in H0.
inversion_clear H0.
inversion_clear H3.
inversion_clear H4.
rewrite H5 in H1.
unfold T_cont_p.
auto.
intros R M M'.
unfold T_loop_p.
unfold T_cont_G.
unfold T_loop_G.
intros.
apply H2; auto.
unfold T_cont_A.
auto.
unfold T_cont_G.
unfold T_loop_G.
intros.
apply H2; auto.
unfold Psi.
unfold get.
auto.
intros R M.
unfold T_loop_p.
omega.
unfold T_G_id.
unfold T_loop_G.
intros.
rewrite H2.
auto.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id.
auto.
intros.
omega.

(* st r0 0 r0 *)
apply I_infer_st with 
  (fun s : state =>
      let (R, _) := s in R r0 >= 0 /\ R r0 < 100 /\ R r1 = 1 /\ R r2 = 100)
  T_G_id T_loop_A T_loop_G.
auto.
intros.
unfold T_G_id in H1.
rewrite <- H1.
unfold T_loop_G.
intros.
unfold ext_w.
unfold ext.
assert ((0+i) <> (R r0 + 0)).
assert (0 + i = i); auto.
rewrite H5.
assert (R r0 + 0 = R r0).
auto.
rewrite H6.
apply lt_neq.
auto.
assert (false = beq_w (0 + i) (R r0 + 0)).
unfold beq_w.
apply beq_neq_false.
auto.
rewrite <- H6.
auto.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.
tauto.

(* fork L_chld r0 *)
apply I_infer_fork with T_chld_p T_fork_A T_chld_A T_fork_G T_chld_G.
unfold T_chld_p.
intros.
unfold ext_r.
unfold ext.
unfold beq_r.
tauto.
unfold T_fork_G.
unfold T_chld_A.
intros.
unfold ext_r.
unfold ext.
unfold beq_r.
assert (0 + (R r0) = R r0); auto.
rewrite <- H3.
inversion_clear H0.
inversion_clear H5.
apply H2; auto.

unfold T_chld_G.
unfold T_fork_A.
intros.
apply H2; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
assert (R r0 <> i).
apply lt_neq; auto.
auto.

unfold T_loop_A.
unfold T_chld_A.
intros.
unfold ext_r.
unfold ext.
unfold beq_r.
assert (0 + (R r0) = R r0).
auto.
rewrite <- H3.
inversion_clear H0.
inversion_clear H5.
apply H2; auto.

unfold T_chld_G.
unfold T_loop_G.
intros.
apply H2; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
apply lt_neq; auto.

unfold T_loop_A.
unfold T_fork_A.
intros.
apply H1; auto.
omega.

unfold T_fork_G.
unfold T_loop_G.
intros.
apply H1; auto.
omega.
unfold T_chld_p.
auto.

unfold Psi.
unfold get.
auto.

(* yield *)
apply I_infer_yield.
unfold T_loop_G.
auto.
tauto.

(* add r0 r0 r1 *)
apply I_infer_add with T_loop_p T_loop_G T_loop_A T_loop_G.
discriminate.
discriminate.
unfold T_loop_p.
unfold ext_r.
unfold ext.
unfold beq_r.
intros.
inversion_clear H0.
inversion_clear H2.
inversion_clear H3.
rewrite H2.
split.
auto with arith.
split.
omega.
split; auto.

unfold T_loop_G.
unfold T_fork_G.
unfold ext_r.
unfold ext.
unfold beq_r.
intros.
apply H1; auto.
omega.

unfold T_fork_A.
unfold T_loop_A.
unfold ext_r.
unfold ext.
unfold beq_r.
intros.
apply H1; omega.

unfold T_loop_G.
unfold T_fork_G.
unfold ext_r.
unfold ext.
unfold beq_r.
intros.
apply H1; auto.
omega.

(* yield *)
apply I_infer_yield.
unfold T_loop_G.
auto.
tauto.

(* jd L_loop *)
apply I_infer_jd with T_loop_p T_loop_G T_loop_A T_loop_G; auto.
unfold Psi.
unfold get.
auto.

Qed.



Lemma wf_cont : I_infer Inv Psi (T_cont_p, T_cont_G, T_cont_A, T_cont_G) C_cont.
Proof.
unfold C_cont.

(* yield *)
apply I_infer_yield.
unfold T_cont_G.
auto.
tauto.

(* jd L_cont *)
apply I_infer_jd with T_cont_p T_cont_G T_cont_A T_cont_G; auto.
unfold Psi.
unfold get.
auto.

Qed.

Lemma wf_chld : I_infer Inv Psi (T_chld_p, T_chld_G, T_chld_A, T_chld_G) C_chld.
Proof.
unfold C_chld.

(* movi *)
apply I_infer_movi with 
  (fun s : state => let (R,M) := s in T_chld_p (R,M) /\ R r0 = V_data) T_chld_G T_chld_A T_chld_G.
discriminate.
discriminate.
unfold T_chld_p.
intros.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* yield *)
apply I_infer_yield.
unfold T_chld_G.
auto.
tauto.

(* add *)
apply I_infer_add with 
  (fun s : state => let (R,M) := s in R r0 = V_data + R ta) T_chld_G T_chld_A T_chld_G.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
intros.
inversion_clear H0.
rewrite H2.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* yield *)
apply I_infer_yield.
unfold T_chld_G.
auto.
tauto.

(* ld *)
apply I_infer_ld with 
  (fun s : state => let (R,M) := s in R r0 = V_data + R ta) T_chld_G T_chld_A T_chld_G.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.


(* yield *)
apply I_infer_yield.
unfold T_chld_G.
auto.
tauto.

(* movi *)
apply I_infer_movi with 
  (fun s : state => let (R,M) := s in R r0 = V_data + R ta) T_chld_G T_chld_A T_chld_G.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* yield *)
apply I_infer_yield.
unfold T_chld_G.
auto.
tauto.

(* add *)
apply I_infer_add with 
  (fun s : state => let (R,M) := s in R r0 = V_data + R ta) T_chld_G T_chld_A T_chld_G.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_chld_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* yield *)
apply I_infer_yield.
unfold T_chld_G.
auto.
tauto.

(* st *)
apply I_infer_st with 
  (fun s : state => True) T_chld_G T_chld_A T_chld_G.
auto.

unfold T_chld_G.
intros.
assert (ext_w M (R r0 + 0) (R r1) (0 + i) = M (0 + i)).
assert (R r0 = R ta).
apply trans_eq with (0 + R ta); auto.
rewrite <- H5 in H4.
assert (R r0 + 0 = R r0).
auto with arith.
rewrite H6.
assert (0 + i = i); auto.
rewrite H7.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (false = beq_nat i (R r0)).
apply beq_neq_false.
auto.
rewrite <- H8.
auto.
rewrite <-H5.
apply H1; auto.

auto.
auto.

(* exit *)
apply I_infer_exit.
unfold T_chld_G; auto.
Qed.

Lemma wf_CdHeap : C_infer Inv Psi CdHeap.
Proof.
apply C_infer_1.
unfold Psi.
unfold CdHeap.
intros f p g A G.
assert (f = 0 \/ f = 1 \/ f = 2 \/ f = 3 \/ f = 4 \/ f > 4).

Lemma cdhp_tmp0 : forall f, f = 0 \/ f = 1 \/ f = 2 \/ f = 3 \/ f = 4 \/ f > 4.
induction f.
left; auto.
inversion_clear IHf.
right.
left.
auto.
inversion_clear H.
right.
right.
left.
auto.
inversion_clear H0.
right.
right.
right.
left.
auto.
inversion_clear H.
right.
right.
right.
right.
left.
auto.
inversion_clear H0.
right.
right.
right.
right.
right.
rewrite H.
auto with arith.

right.
right.
right.
right.
right.
auto with arith.
Qed.
apply cdhp_tmp0.

unfold get.
unfold beq_w.
inversion_clear H.
rewrite H0.
unfold beq_nat.
intros.
inversion_clear H.

inversion_clear H0.
rewrite H.
unfold beq_nat.
intros.
exists C_main.
auto.

inversion_clear H.
rewrite H0.
unfold beq_nat.
intros.
exists C_loop.
auto.

inversion_clear H0.
rewrite H.
unfold beq_nat.
intros.
exists C_cont.
auto.

inversion_clear H.
rewrite H0.
unfold beq_nat.
intros.
exists C_chld.
auto.

assert (4 <> f).
apply lt_neq; auto.
assert (3 <> f).
apply lt_neq.
apply lt_trans with 4; auto.
assert (2 <> f).
apply lt_neq.
apply lt_trans with 4; auto.
assert (1 <> f).
apply lt_neq.
apply lt_trans with 4; auto.

Lemma cdhp_tmp1 : forall a b, a <> b -> false = beq_nat b a.
intros.
apply beq_neq_false; auto.
Qed.

assert (false = beq_nat f 1).
apply cdhp_tmp1; auto.

assert (false = beq_nat f 2).
apply cdhp_tmp1; auto.
assert (false = beq_nat f 3).
apply cdhp_tmp1; auto.
assert (false = beq_nat f 4).
apply cdhp_tmp1; auto.
rewrite <- H4.
rewrite <- H5.
rewrite <- H6.
rewrite <- H7.
intro.
inversion_clear H8.

intros f I p g A G.
intro.
assert (f = 0 \/ f = 1 \/ f = 2 \/ f = 3 \/ f = 4 \/ f > 4).
apply cdhp_tmp0.
unfold Psi in H.
unfold CdHeap in H.
unfold get in H.
unfold beq_w in H.
inversion_clear H0.
rewrite H1 in H.
unfold beq_nat in H.
inversion_clear H.
inversion_clear H2.

inversion_clear H1.
rewrite H0 in H.
unfold beq_nat in H.
inversion_clear H.
inversion_clear H1.
inversion_clear H2.
apply wf_main.

inversion_clear H0.
rewrite H1 in H.
unfold beq_nat in H.
inversion_clear H.
inversion_clear H0.
inversion_clear H2.
apply wf_loop.

inversion_clear H1.
rewrite H0 in H.
unfold beq_nat in H.
inversion_clear H.
inversion_clear H1.
inversion_clear H2.
apply wf_cont.

inversion_clear H0.
rewrite H1 in H.
unfold beq_nat in H.
inversion_clear H.
inversion_clear H0.
inversion_clear H2.
apply wf_chld.

assert (4 <> f).
apply lt_neq; auto.
assert (3 <> f).
apply lt_neq.
apply lt_trans with 4; auto.
assert (2 <> f).
apply lt_neq.
apply lt_trans with 4; auto.
assert (1 <> f).
apply lt_neq.
apply lt_trans with 4; auto.

assert (false = beq_nat f 1).
apply cdhp_tmp1; auto.

assert (false = beq_nat f 2).
apply cdhp_tmp1; auto.
assert (false = beq_nat f 3).
apply cdhp_tmp1; auto.
assert (false = beq_nat f 4).
apply cdhp_tmp1; auto.
rewrite <- H8 in H.
rewrite <- H5 in H.
rewrite <- H6 in H.
rewrite <- H7 in H.
inversion_clear H.
inversion_clear H10.
Qed.

Definition M0 : heap := fun l : nat => 0.
Definition R0 : regs := 
  fun r : reg => 
    if beq_r r tn
    then 1
    else 0.
Definition Q0 : queue :=
  fun t : tid => NoneT.
Definition Theta0 : qspec :=
  fun t: tid => NoneT.

Theorem wf_prog : P_infer (Inv, Psi, Theta0, (T_main_p, T_main_G, T_main_A, T_main_G)) (CdHeap, (R0,M0), Q0, C_main).
Proof.
apply P_infer_1.
apply wf_CdHeap.

(* Q_infer *)
apply Q_infer_1.
intros.
unfold Q0 in H.
unfold get in H.
inversion_clear H.

exists 1.
intros.
unfold Q0 in H.
unfold has in H.
inversion_clear H.
inversion_clear H0.

intros.
unfold Q0 in H.
unfold get in H.
inversion_clear H.

intros.
inversion_clear H.
unfold Q0 in H1.
unfold get in H1.
inversion_clear H1.

intros.
inversion_clear H.
unfold Q0 in H4.
unfold get in H4.
inversion_clear H4.

apply Q_nonin_1.
intros.
inversion_clear H.
inversion_clear H2.
unfold Theta0 in H.
unfold get in H.
inversion_clear H.

apply wf_main.
unfold Inv. auto.
unfold T_main_p.
auto.

intro.
unfold Q0 in H.
unfold has in H.
inversion_clear H.
inversion_clear H0.

intros.
inversion_clear H.
unfold Q0 in H2.
unfold get in H2.
inversion_clear H2.

apply P_nonin_1.
intros.
inversion_clear H.
unfold Q0 in H2.
unfold get in H2.
inversion_clear H2.

intros.
inversion_clear H.
unfold Q0 in H2.
unfold get in H2.
inversion_clear H2.
Qed.

